 /*
  * sigact.c
  * Copyright (C) Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
  * 
  * simple-netaid is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  * 
  * simple-netaid is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  * See the GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  * 
  * See the COPYING file.
  */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>

#include <cdk_test.h>

#include <simple-netaid/entry.h>

#include "sigact.h"
#include <globals.h>

struct sigaction sa;

static
void hdl(int signum, siginfo_t *info, void *ucontext)
{
   switch(signum)
   {
      case SIGALRM:
         sigaction_init();
         break;
      case SIGCHLD:
         break;
      case SIGHUP:
         break;
      case SIGUSR1:
         sigaction_init();
         break;
      case SIGUSR2:
         sigaction_init();
         break;
      case SIGTERM:
      case SIGINT:
      case SIGQUIT:
         /* Make sure to cleanup if user sends a keyboard interrupt */
         ui_cleanup();
         ui_delete();
         endCDK();
         emit_sigusr(-1, libpath);
         exit(EXIT_SUCCESS);
   }
}

void sigaction_init()
{
   memset (&sa, 0, sizeof(sa));
      
   /*---------------- initialize signal handling ------------ */
   //sa.sa_handler = hdl;   
   sa.sa_sigaction = &hdl;
   sa.sa_flags = SA_RESTART; // | SA_SIGINFO;
   sigemptyset(&sa.sa_mask);
   sigaction(SIGALRM, &sa, 0);
   sigaction(SIGCHLD, &sa, 0);
   sigaction(SIGHUP,  &sa, 0);
   sigaction(SIGUSR1, &sa, 0);
   sigaction(SIGUSR2, &sa, 0);
   sigaction(SIGTERM, &sa, 0);
   sigaction(SIGINT,  &sa, 0);
   sigaction(SIGQUIT, &sa, 0);
}

void wait_for_sigusr1()
{
   sigset_t mask;
   sigemptyset(&mask);
   sigaddset(&mask, SIGUSR1);
   sigprocmask(SIG_BLOCK, &mask, NULL);

   if (sigprocmask(SIG_BLOCK, &mask, NULL) == -1) {
      fprintf(stderr, "Cannot block SIGUSR1: %s.\n", strerror(errno));
      exit(EXIT_FAILURE);
   }
       
   while (1) {  
      siginfo_t info;
      int signum = sigwaitinfo(&mask, &info);
      if (signum == -1) {		 
         // If some other signal was delivered to a handler installed
         // without SA_RESTART in sigaction flags, it will interrupt
         // slow calls like sigwaitinfo() with EINTR error. So, those
         // are not really errors.
         if (errno == EINTR) {
            continue;
         }
         
      } else if (signum == SIGUSR1) {
         char *c = NULL;
         if (asprintf(&c, "%jd", (intmax_t) info.si_pid) != -1) {
            char output[1024];
            FILE *pf = popen("ps -A | grep \"snetaid$\"", "r");
            if(!pf) {
               fprintf(stderr, "Pipe error %s.\n", strerror(errno));
               exit(EXIT_FAILURE);
            } 
              
            fgets(output, 1024, pf);
            pclose(pf);
            char *token = strtok((char*)output, " ");
            if(!strcmp(c, token)) {   
               free(c); // cleanup when done.
               break;
            }
         }             
      } else if (signum == SIGINT) {     // Also exit if Ctrl+C pressed in terminal (SIGINT).
         printf("Ctrl+C pressed.\n");
         break;
      } else {      
         printf("Unable to get the pid.\n");
      }
   
   }  // while
   
   sigdelset(&mask, SIGUSR1);
}
